import tflearn
import numpy as np
import tensorflow as tf
import pickle
from scipy.special import expit
import glob
from sklearn.model_selection import train_test_split

def trainDualStreamRNN(testing=False):
    """

        Trains a Dual-Stream Architecture from scratch.


        1. Load dataset
        2. Define Network Architecture
        3. Define loss and optimizer
        4. Start training
        5. Store best performing model


        :return: Nothing
        """

    # How many frames we are using for our sequence
    numFrames = 20
    numFeatures = 512

    # Load temporal training data
    with open(
            '/home/haakon/PycharmProjects/DNN-Masters-code/RNN/Vectors/fromVideoVectors/ResNet34_Type:A_KeepTimeDim(1, 0, 0, 0)_TimeStep=20_GlobalAvgPool_trainBatchX',
            'rb') as fp:
        X_batchTemporal = pickle.load(fp)

    with open(
            '/home/haakon/PycharmProjects/DNN-Masters-code/RNN/Vectors/fromVideoVectors/ResNet34_Type:A_KeepTimeDim(1, 0, 0, 0)_TimeStep=20_GlobalAvgPool_trainBatchY',
            'rb') as fp:
        Y_batchTemporal = pickle.load(fp)

    # Load spatial training data

    with open(
            '/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/DistributionSamples/Spatial/fc7_trainBatchX',
            'rb') as fp:
        X_batchSpatial = pickle.load(fp)

    with open(
            '/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/DistributionSamples/Spatial/fc7_trainBatchY',
            'rb') as fp:
        Y_batchSpatial = pickle.load(fp)

    # Load temporal validation data
    with open(
            '/home/haakon/PycharmProjects/DNN-Masters-code/RNN/Vectors/fromVideoVectors/ResNet34_Type:A_KeepTimeDim(1, 0, 0, 0)_TimeStep=20_GlobalAvgPool_valBatchX',
            'rb') as fp:
        X_batch_ValTemporal = pickle.load(fp)

    with open(
            '/home/haakon/PycharmProjects/DNN-Masters-code/RNN/Vectors/fromVideoVectors/ResNet34_Type:A_KeepTimeDim(1, 0, 0, 0)_TimeStep=20_GlobalAvgPool_valBatchY',
            'rb') as fp:
        Y_batch_ValTemporal = pickle.load(fp)

    # Load spatial valdiation data
    with open(
            '/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/DistributionSamples/Spatial/fc7_valBatchX',
            'rb') as fp:
        X_batch_ValSpatial = pickle.load(fp)

    with open(
            '/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/DistributionSamples/Spatial/fc7_valBatchY',
            'rb') as fp:
        Y_batch_ValSpatial = pickle.load(fp)

    # Input placeholders
    x1 = tf.placeholder(shape=(None, numFrames, 512), dtype=tf.float32, name='x_input')
    x2 = tf.placeholder(shape=(None, numFrames, 4096), dtype=tf.float32, name='x_input')
    y_ = tf.placeholder(shape=(None, 1),  dtype=tf.float32, name = 'y_correct')

    # Network input
    net1 = tflearn.input_data(placeholder=x1)
    net2 = tflearn.input_data(placeholder=x2)

    # Merge input vectors and concatenate
    net = tflearn.merge([net1, net2], 'concat', axis=2)

    # Batch Normalize input
    net = tflearn.batch_normalization(net)

    # Build LSTM network
    lstm = tflearn.lstm(net, 256, dropout=0.5)

    # Net output
    net = tflearn.fully_connected(lstm, 1, name='Net_out')

    # Define loss
    loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=y_, logits=net))

    # Define train optimizer
    train_op = tf.train.AdamOptimizer(learning_rate=0.000001).minimize(loss)

    # Define accuracy calculation
    correct_prediction = tf.equal(tf.round(tf.nn.sigmoid(net)), y_)
    accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))

    # Initialize variables
    init = tf.global_variables_initializer()


    num_epochs = 50
    batchSize = 10
    trainIter = 0
    valIter = 0
    testIter = 0
    bestVal = 0

    saver = tf.train.Saver()

    modelname = "OneLayer_Combined_BiLSTM_BatchNorm@Start_Dropout@End_SequenceLength:" + str(numFrames)

    # Start session
    with tf.Session() as sess:
        sess.run(init)

        # Run for num epochs
        for epoch_number in range(num_epochs):
            trainStartPos = 0
            valStartPos = 0
            testStartPos = 0
            totalVal = 0
            totalTest = 0
            trainEndPos = batchSize
            valEndPos = batchSize
            testEndPos = batchSize

            # Shuffle every time
            idx = np.random.permutation(X_batchTemporal.shape[0])
            X_batchTemporal = X_batchTemporal[idx, :, :]
            X_batchSpatial = X_batchSpatial[idx, :, :]
            Y_batchTemporal = Y_batchTemporal[idx, :]

            # Turn on training parameters
            tflearn.is_training(True)

            # Run mini-batches
            for step in range(len(X_batchTemporal)/batchSize):
                # load mini-batches
                batch_x1 = X_batchTemporal[trainStartPos:trainEndPos]
                batch_x2 = X_batchSpatial[trainStartPos:trainEndPos]
                batch_y = Y_batchTemporal[trainStartPos:trainEndPos]

                # Do forward and backward pass with mini-batch, updating the weights
                _, loss_train, acc_train, pred_train, LSTM = sess.run([train_op, loss, accuracy, net, lstm],
                                                                feed_dict={x1: batch_x1, x2: batch_x2, y_: batch_y})

                print LSTM.shape

                trainStartPos+=batchSize
                trainEndPos+=batchSize

                # Print results
                print("Epoch: " + str(epoch_number) \
                      + ", Step: " + str(trainIter) \
                      + " --Loss: " + str("%.4f" % round(loss_train, 4)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_train, 4)))


                trainIter += 1

                # Print comparrison between net output and true labels
                if (trainIter % 100) == 0:
                    print "Accuracy: " + str(acc_train)
                    print 'Correct \t\tPred'
                    for n, g in zip(batch_y, pred_train):
                        print str(n) + '\t\t\t' + str(expit(g))


            # Validation

            # Turn off training parameters
            tflearn.is_training(False)

            # Run all mini-batches in validation data
            for val_step in range(len(X_batch_ValTemporal) / batchSize):
                # Get new batches
                val_batch_x1 = X_batch_ValTemporal[valStartPos:valEndPos]
                val_batch_x2 = X_batch_ValSpatial[valStartPos:valEndPos]
                val_batch_y = Y_batch_ValTemporal[valStartPos:valEndPos]

                # Perform forward pass WITHOUT the backward pass
                loss_val, acc_val, pred_val = sess.run([loss, accuracy, net],
                                                       feed_dict={x1: val_batch_x1, x2: val_batch_x2, y_: val_batch_y})

                totalVal += acc_val

                valStartPos += batchSize
                valEndPos += batchSize

                # Print results from forward pass
                print("---- VALIDATION -------"
                      + " Step: " + str(valIter) \
                      + " --Loss: " + str("%.4f" % round(loss_val, 4)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_val, 4)))

                # Print comparrison between net output and true labels
                if val_step % 5 == 0:
                    print "Accuracy: " + str(acc_val)
                    print 'Correct \t\tPred'
                    for n, g in zip(val_batch_y, pred_val):
                        print str(n) + '\t\t\t' + str(expit(g))

                valIter += 1
            # Calculate total validation performance
            totalValidation = float(totalVal) / float(val_step + 1)
            VALIDATION = str("%.4f" % round(totalValidation, 4))
            print "Total validation accuracy: " + VALIDATION

            # If new best validation results, save model
            if VALIDATION > bestVal:
                bestVal = VALIDATION
                save_path = saver.save(sess, "Models/Combined/" + modelname + "_Val_"
                                       + VALIDATION + "_step", global_step=trainIter)



        if testing:
            print "TESTING!"
            # TESTING
            #   feeding temporal
            testfeedvectorTemp_paths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Temporal/videoVectors/Testing/feeding/Vectors/GlobalAvgPool/*")
            testfeedvectorTemp_paths.sort(key=lambda f: int(filter(str.isdigit, f)))
            testfeedlabelsTemp_paths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Temporal/videoVectors/Testing/feeding/Labels/*")
            testfeedlabelsTemp_paths.sort(key=lambda f: int(filter(str.isdigit, f)))

            #   feeding spatial
            testfeedvectorSpat_paths = glob.glob(
                "/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/Every20thSpatialVector/Testing/feeding/Vectors/fc7/*")
            testfeedvectorSpat_paths.sort(key=lambda f: int(filter(str.isdigit, f)))
            testfeedlabelsSpat_paths = glob.glob(
                "/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/Every20thSpatialVector/Testing/feeding/Labels/fc7/*")
            testfeedlabelsSpat_paths.sort(key=lambda f: int(filter(str.isdigit, f)))

            #   nonfeeding temporal
            testNonfeedvectorTemp_paths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Temporal/videoVectors/Testing/nonfeeding/Vectors/GlobalAvgPool/*")
            testNonfeedvectorTemp_paths.sort(key=lambda f: int(filter(str.isdigit, f)))
            testNonfeedlabelsTemp_paths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Temporal/videoVectors/Testing/nonfeeding/Labels/*")
            testNonfeedlabelsTemp_paths.sort(key=lambda f: int(filter(str.isdigit, f)))

            #   nonfeeding spatial
            testNonfeedvectorSpat_paths = glob.glob(
                "/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/Every20thSpatialVector/Testing/nonfeeding/Vectors/fc7/*")
            testNonfeedvectorSpat_paths.sort(key=lambda f: int(filter(str.isdigit, f)))
            testNonfeedlabelsSpat_paths = glob.glob(
                "/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/Every20thSpatialVector/Testing/nonfeeding/Labels/fc7/*")
            testNonfeedlabelsSpat_paths.sort(key=lambda f: int(filter(str.isdigit, f)))


            # Test vectors
            test_vectors_temp = []
            test_vectors_spat = []
            test_labels = []
            for path in testfeedvectorTemp_paths:
                with open(path, 'rb') as fp:
                    test_vectors_temp += pickle.load(fp)

            for path in testfeedvectorSpat_paths:
                with open(path, 'rb') as fp:
                    test_vectors_spat += pickle.load(fp)

            for path in testfeedlabelsTemp_paths:
                with open(path, 'rb') as fp:
                    test_labels += pickle.load(fp)

            for path in testNonfeedvectorTemp_paths:
                with open(path, 'rb') as fp:
                    test_vectors_temp += pickle.load(fp)

            for path in testNonfeedvectorSpat_paths:
                with open(path, 'rb') as fp:
                    test_vectors_spat += pickle.load(fp)

            for path in testNonfeedlabelsTemp_paths:
                with open(path, 'rb') as fp:
                    test_labels += pickle.load(fp)
            print len(test_vectors_temp)
            print len(test_vectors_spat)

            # Test batches
            X_batchTemp_test = np.zeros(shape=[len(test_vectors_temp) / numFrames, numFrames, 512])
            X_batchSpat_test = np.zeros(shape=[len(test_vectors_spat) / numFrames, numFrames, 4096])
            Y_batch_test = np.zeros(shape=[len(test_vectors_temp) / numFrames, 1])

            startPos = 0
            endPos = startPos + numFrames

            for i in range(len(X_batchTemp_test)):
                temp_batch = np.asarray(test_vectors_temp[startPos:endPos][:])
                spat_batch = np.asarray(test_vectors_spat[startPos:endPos][:])
                X_batchTemp_test[i] = temp_batch
                X_batchSpat_test[i] = spat_batch
                Y_batch_test[i] = test_labels[startPos]
                startPos += numFrames
                endPos += numFrames

            print len(X_batchTemp_test)

            tflearn.is_training(False)
            for test_step in range(len(X_batchTemp_test) / batchSize):
                test_batch_x1 = X_batchTemp_test[testStartPos:testEndPos]
                test_batch_x2 = X_batchSpat_test[testStartPos:testEndPos]
                test_batch_y = Y_batch_test[testStartPos:testEndPos]
                loss_test, acc_test, pred_test = sess.run([loss, accuracy, net],
                                                               feed_dict={x1: test_batch_x1, x2: test_batch_x2, y_: test_batch_y})

                totalTest += acc_test

                testStartPos += batchSize
                testEndPos += batchSize

                print("---- TESTING -------"
                        + " Step: " + str(testIter) \
                        + " --Loss: " + str("%.4f" % round(loss_test, 4)) \
                        + "--  --Accuracy: " + str("%.4f" % round(acc_test, 4)))

                # Print comparrison between net output and true labels
                if test_step % 1 == 0:
                    print "Accuracy: " + str(acc_test)
                    print 'Correct \t\tPred'
                    for n, g in zip(test_batch_y, pred_test):
                        print str(n) + '\t\t\t' + str(expit(g))

                testIter += 1
            totalTest = float(totalTest) / float(test_step + 1)
            TEST = str("%.4f" % round(totalTest, 4))
            print "Total TEST accuracy: " + TEST


            print 'Best Validation result = ' + str(bestVal)

#trainDualStreamRNN(testing='False')