import tflearn
import numpy as np
import tensorflow as tf
import pickle
from scipy.special import expit
from sklearn.model_selection import train_test_split
import glob

def trainSpatialRNN():
    """

    Trains a Spatial Recurrent Architecture from scratch.


    1. Load dataset
    2. Define Network Architecture
    3. Define loss and optimizer
    4. Start training
    5. Store best performing model


    :return: Nothing
    """

    # Load training data paths
    trainVecFeedPaths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/videoVectors/Training/feeding/Vectors/fc7/*")
    trainLabFeedPaths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/videoVectors/Training/feeding/Labels/*")

    trainVecNonFeedPaths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/videoVectors/Training/nonfeeding/Vectors/fc7/*")
    trainLabNonFeedPaths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/videoVectorsTraining/nonfeeding/Labels/*")



    # Load validation data paths
    valVecFeedPaths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/videoVectors/Validating/feeding/Vectors/fc7/*")
    valLabFeedPaths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/videoVectors/Validating/feeding/Labels/*")

    valVecNonFeedPaths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/videoVectors/Validating/nonfeeding/Vectors/fc7/*")
    valLabNonFeedPaths = glob.glob("/media/haakon/Storage/hakon/Documents/Masters/Dataset/Spatial/videoVectors/Validating/nonfeeding/Labels/*")



    trainfeed_vectors = []
    trainnonfeed_vectors = []
    trainfeed_labels = []
    trainnonfeed_labels = []

    valfeed_vectors = []
    valnonfeed_vectors = []
    valfeed_labels = []
    valnonfeed_labels = []

    # Load training vectors
    for path in trainVecFeedPaths:
        with open (path, 'rb') as fp:
            trainfeed_vectors += pickle.load(fp)

    for path in trainLabFeedPaths:
        with open (path, 'rb') as fp:
            trainfeed_labels += pickle.load(fp)

    for path in trainVecNonFeedPaths:
        with open(path, 'rb') as fp:
            trainnonfeed_vectors += pickle.load(fp)

    for path in trainLabNonFeedPaths:
        with open(path, 'rb') as fp:
            trainnonfeed_labels += pickle.load(fp)

    # Load Validation vectors
    for path in valVecFeedPaths:
        with open(path, 'rb') as fp:
            valfeed_vectors += pickle.load(fp)

    for path in valLabFeedPaths:
        with open(path, 'rb') as fp:
            valfeed_labels += pickle.load(fp)

    for path in valVecNonFeedPaths:
        with open(path, 'rb') as fp:
            valnonfeed_vectors += pickle.load(fp)

    for path in valLabNonFeedPaths:
        with open(path, 'rb') as fp:
            valnonfeed_labels += pickle.load(fp)


    numFrames = 20
    numFeatures = 4096


    # Input placeholders
    x = tf.placeholder(shape=(None, numFrames, numFeatures), dtype=tf.float32, name='x_input')
    y_ = tf.placeholder(shape=(None, 1),  dtype=tf.float32, name = 'y_correct')

    # Network input
    net = tflearn.input_data(placeholder=x)

    # Batch Normalize input
    net = tflearn.batch_normalization(net)

    # Build LSTM network
    net = tflearn.lstm(net, 256, dropout=0.5)

    # Net output
    net = tflearn.fully_connected(net, 1, name='Net_out')

    # Define loss
    loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=y_, logits=net))

    # Define train optimizer
    train_op = tf.train.AdamOptimizer(learning_rate=0.000001).minimize(loss)

    # Define accuracy calculation
    correct_prediction = tf.equal(tf.round(tf.nn.sigmoid(net)), y_)
    accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))

    # Initialize variables
    init = tf.global_variables_initializer()


    num_epochs = 50
    batchSize = 10
    trainIter = 0
    valIter = 0
    bestVal = 0

    # Create set of indices spaced evenly appart at numFrames length
    selectionFeedVectors = np.arange(0, len(trainfeed_vectors)-numFrames, numFrames)
    selectionNonFeedVectors = np.arange(0, len(trainnonfeed_vectors)-numFrames, numFrames)

    # Select distribution
    selectionNonFeedVectors, _ = train_test_split(selectionNonFeedVectors, train_size=0.75, random_state=42)

    # Create set of indices spaced evenly appart at numFrames length
    selectionValFeedVectors = np.arange(0, len(valfeed_vectors)-numFrames, numFrames)
    selectionValNonFeedVectors = np.arange(0, len(valnonfeed_vectors)-numFrames, numFrames)

    # Select distribution
    selectionValNonFeedVectors, _ = train_test_split(selectionValNonFeedVectors, train_size=0.75, random_state=42)
    saver = tf.train.Saver()

    modelname = "OneLayer_SPATIAL_VanillaRNN_BatchNorm@Start_Dropout@End_SequenceLength:" + str(numFrames)

    # Start session
    with tf.Session() as sess:
        sess.run(init)
        batch_x = np.zeros(shape=[batchSize, numFrames, numFeatures])
        batch_y = np.zeros(shape=[batchSize, 1])

        val_batch_x = np.zeros(shape=[batchSize, numFrames, numFeatures])
        val_batch_y = np.zeros(shape=[batchSize, 1])

        # Run for num epochs
        for epoch_number in range(num_epochs):
            frames = np.zeros(shape=[numFrames,numFeatures])
            labels = np.zeros(shape=[numFrames,1])

            # Shuffle every time
            idxfeed = np.random.permutation(selectionFeedVectors.shape[0])
            idxnonfeed = np.random.permutation(selectionNonFeedVectors.shape[0])

            selectionFeedVectors = selectionFeedVectors[idxfeed]
            selectionNonFeedVectors = selectionNonFeedVectors[idxnonfeed]

            # Turn on training parameters
            tflearn.is_training(True)

            feediter = 0
            nonfeediter = 0

            # Run mini-batches
            for step in range(len(selectionFeedVectors*2)/batchSize):
                totalVal = 0


                for i in range(batchSize):
                    # Create mini-batch
                    if i < batchSize/2:
                        for j in range(numFrames):
                            currentFrame = trainfeed_vectors[selectionFeedVectors[feediter]+j]
                            frames[j] = currentFrame

                        feediter+=1
                        batch_x[i] = frames
                        batch_y[i] = 1
                    else:
                        for k in range(numFrames):
                            currentFrame = trainnonfeed_vectors[selectionNonFeedVectors[nonfeediter]+k]
                            frames[k] = currentFrame

                        nonfeediter+=1
                        batch_x[i] = frames
                        batch_y[i] = 0


                # Do forward and backward pass with mini-batch, updating the weights
                _, loss_train, acc_train, pred_train = sess.run([train_op, loss, accuracy, net],
                                                                feed_dict={x: batch_x, y_: batch_y})



                # Print results
                print("Epoch: " + str(epoch_number) \
                      + ", Step: " + str(trainIter) \
                      + " --Loss: " + str("%.4f" % round(loss_train, 4)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_train, 4)))


                trainIter += 1

                # Print comparrison between net output and true labels
                if (trainIter % 100) == 0:
                    print "Accuracy: " + str(acc_train)
                    print 'Correct \t\tPred'
                    for n, g in zip(batch_y, pred_train):
                        print str(n) + '\t\t\t' + str(expit(g))


            # Validation

            # Turn off training parameters
            tflearn.is_training(False)
            feediter = 0
            nonfeediter = 0

            for val_step in range(len(selectionValFeedVectors*2) / batchSize):


                for i in range(batchSize):
                    # Get new batches
                    if i < batchSize / 2:
                        for j in range(numFrames):
                            frames[j] = valfeed_vectors[selectionValFeedVectors[feediter]+j]
                        feediter += 1
                        val_batch_x[i] = frames
                        val_batch_y[i] = 1
                    else:
                        for j in range(numFrames):
                            frames[j] = valnonfeed_vectors[selectionValNonFeedVectors[nonfeediter]+j]
                        nonfeediter += 1
                        val_batch_x[i] = frames
                        val_batch_y[i] = 0

                # Perform forward pass WITHOUT the backward pass
                loss_val, acc_val, pred_val = sess.run([loss, accuracy, net],
                                                       feed_dict={x: val_batch_x, y_: val_batch_y})

                totalVal += acc_val

                # Print results from forward pass
                print("---- VALIDATION -------"
                      + " Step: " + str(valIter) \
                      + " --Loss: " + str("%.4f" % round(loss_val, 4)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_val, 4)))

                # Print comparrison between net output and true labels
                if val_step % 5 == 0:
                    print "Accuracy: " + str(acc_val)
                    print 'Correct \t\tPred'
                    for n, g in zip(val_batch_y, pred_val):
                        print str(n) + '\t\t\t' + str(expit(g))

                valIter += 1

            # Calculate total validation performance
            totalValidation = float(totalVal) / float(val_step + 1)
            VALIDATION = str("%.4f" % round(totalValidation, 4))
            print "Total validation accuracy: " + VALIDATION

            # If new best validation results, save model
            if VALIDATION > bestVal:
                bestVal = VALIDATION
                save_path = saver.save(sess, "Models/" + modelname + "_Val_"
                                       + VALIDATION + "_step", global_step=trainIter)

        print 'Best Validation result = ' + bestVal

#trainSpatialRNN()