import glob
import random

import cv2
import numpy as np
from PIL import Image

from imageProcessing import zeroCenterOFData, zeroCenterGrayData



class File_Handler(object):
    _original_feed_list = []
    _original_nonfeed_list = []
    _timeStep_feed_list = []
    _timeStep_nonfeed_list = []
    _numTimesteps = 0
    _datasetType = 0

    def __init__(self, feed_folder_path, dataType, nonfeed_folder_path, numTimesteps, datasetType):
        self._original_feed_list = []
        self._original_nonfeed_list = []
        self._original_feed_list = self.read_image_list(feed_folder_path)
        self._original_nonfeed_list = self.read_image_list(nonfeed_folder_path)
        self._numTimesteps = numTimesteps
        if dataType == "Optical Flow":
            self.populate_timeStep_lists()
        self._datasetType = datasetType




    def read_image_list(self, folder_path):
        """
        Read image paths from the folder path

        :param folder_path: String
            Path to the folder containing the images
        :return: list of image paths
        """
        image_list = glob.glob(folder_path + "/*.jpg")
        image_list.sort(key=lambda f: int(filter(str.isdigit, f)))

        return image_list

    def shuffle_GRAY_data(self):
        """
        Shuffles the lists to make them ready for the next batch

        :return: NOTHING
        """
        random.shuffle(self._original_feed_list)
        random.shuffle(self._original_nonfeed_list)


    def get_GRAY_Batch(self, batch_size, startPos):
        """
        Builds a new mini-batch through selecting images from a list of video frame paths.

        Each mini-batch consists of an equal amount of Feeding and NonFeeding frames.

        Each image is preprocessed using the Specular Removal Preprocesing strategy.


        :param batch_size: size for the mini-batch
        :param startPos: the start position for the list of video frames
        :return: A preprocessed mini-batch, labels and the new start position
        """
        feedImages = np.zeros((batch_size/2) * 224 * 224 * 3, dtype=np.float32)
        feedImages = np.reshape(feedImages, (batch_size/2, 224, 224, 3))

        nonfeedImages = np.zeros(batch_size/2 * 224 * 224 * 3, dtype=np.float32)
        nonfeedImages = np.reshape(nonfeedImages, (batch_size/2, 224, 224, 3))

        # print len(paths)
        for i in range(batch_size / 2):
            # Read in images
            #print self.original_feed_list[startPos]
            feedImg = Image.open(self._original_feed_list[startPos])
            nonfeedImg = Image.open(self._original_nonfeed_list[startPos])


            # Zero center and normalize the images
            feedImages[i] = zeroCenterGrayData(feedImg)
            nonfeedImages[i] = zeroCenterGrayData(nonfeedImg)
            startPos += 1


        feedLabels = np.zeros(shape=(batch_size / 2, 1))
        nonfeedLabels = np.zeros(shape=(batch_size / 2, 1))
        for k in range(batch_size / 2):
            feedLabels[k] = [1]
            nonfeedLabels[k] = [0]

        imageBatch = np.concatenate((feedImages, nonfeedImages))
        labelsBatch = np.concatenate((feedLabels, nonfeedLabels))

        return imageBatch, labelsBatch, startPos


