


from time import gmtime, strftime

import numpy as np
import tensorflow as tf
import tflearn
from tflearn.data_preprocessing import ImagePreprocessing
from scipy.special import expit

from Spatial_File_Handler import File_Handler

def trainSpatialStream():

    """
    Fine tunes an improved VGG-16 from tflearn checkpoint.

    1. Define learning rate and epsilon
    2. Define Network Architecture
    3. Define loss and optimizer
    4. Load checkpoint weights
    5. Load dataset
    6. Start training
    7. Store best performing model


    :return: Nothing
    """
    learning_rate = 0.0001
    epsilon = 1e-8
    testConvergence = False


    # VGG16 preprocessing
    #img_prep = ImagePreprocessing()
    # Remove mean
    #img_prep.add_featurewise_zero_center(mean=[123.68, 116.779, 103.939],
    #                                    per_channel=True)

    # Augment images for training
    #img_aug = tflearn.data_augmentation.ImageAugmentation()
    #img_aug.add_random_flip_leftright()
    #img_aug.add_random_rotation(max_angle=25.)
    #img_aug.add_random_blur(sigma_max=4)

    # Input data placeholders
    x = tf.placeholder(shape=(None, 224, 224, 3), dtype=tf.float32, name='x_placeholder')
    y_ = tf.placeholder(shape=(None, 1),  dtype=tf.float32, name='y_placeholder')




    # Network input
    network = tflearn.input_data(placeholder=x, data_preprocessing=img_prep, name='Net_in')

    #Conv layers
    network = tflearn.conv_2d(network, 64, 3, activation='relu', name='conv1_1')
    network = tflearn.conv_2d(network, 64, 3, activation='relu', name='conv1_2')
    network = tflearn.max_pool_2d(network, 2, strides=2, name='maxpool1')

    network = tflearn.conv_2d(network, 128, 3, activation='relu', name='conv2_1')
    network = tflearn.conv_2d(network, 128, 3, activation='relu', name='conv2_2')
    network = tflearn.max_pool_2d(network, 2, strides=2, name='maxpool2')

    network = tflearn.conv_2d(network, 256, 3, activation='relu', name='conv3_1')
    network = tflearn.conv_2d(network, 256, 3, activation='relu', name='conv3_2')
    network = tflearn.conv_2d(network, 256, 3, activation='relu', name='conv3_3')
    network = tflearn.max_pool_2d(network, 2, strides=2, name='maxpool3')

    network = tflearn.conv_2d(network, 512, 3, activation='relu', name='conv4_1')
    network = tflearn.conv_2d(network, 512, 3, activation='relu', name='conv4_2')
    network = tflearn.conv_2d(network, 512, 3, activation='relu', name='conv4_3')
    network = tflearn.max_pool_2d(network, 2, strides=2, name='maxpool4')

    network = tflearn.conv_2d(network, 512, 3, activation='relu', name='conv5_1')
    network = tflearn.conv_2d(network, 512, 3, activation='relu', name='conv5_2')
    network = tflearn.conv_2d(network, 512, 3, activation='relu', name='conv5_3')
    network = tflearn.max_pool_2d(network, 2, strides=2, name='maxpool5')

    # Batch Normalized Fully connected layers
    network = tflearn.batch_normalization(network, restore=False)
    network = tflearn.fully_connected(network, 4096, activation='relu', name='fc6')

    network = tflearn.batch_normalization(network, restore=False)
    network = tflearn.fully_connected(network, 4096, activation='relu', name='fc7')

    network = tflearn.batch_normalization(network, restore=False)
    network = tflearn.fully_connected(network, 1, restore=False, name='Net_out')







    # Calculate loss
    loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=y_, logits=network))

    # Set up optimizer
    train_op = tf.train.AdamOptimizer(learning_rate=learning_rate, epsilon=epsilon).minimize(loss)

    # Calculate accuracy
    correct_prediction = tf.equal(tf.round(tf.nn.sigmoid(network)), y_)
    accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))


    # Create a summary to monitor cost tensor
    tf.summary.scalar("Pretrained_loss", loss)

    # Create a summary to monitor accuracy tensor
    tf.summary.scalar("Pretrained_accuracy", accuracy)

    # Merge summaries
    merged_summary_op = tf.summary.merge_all()

    # Define regression(necessary for loading TFLearn models)
    modelbuilder = tflearn.regression(network, optimizer='adam',
                                      loss=loss, restore=False)

    # Define model
    model = tflearn.DNN(modelbuilder)

    # Load pretrained model
    model.load('vgg16.tflearn', weights_only=True)




    # Define number of epochs we want to run
    numEpochs = 3
    batch_size = 20

    # Initializing the variables
    init = tf.global_variables_initializer()
    logs_path = 'logs/tensorflowVGG16/Number_of_frames'
    modelName = 'Every_Frame_Straight_Model_LearningRate:'+str(learning_rate)+'_Epsilon:'+str(epsilon)+'_BatchNorm@None_DataAug=None_DataPrep=VGG16-Mean'
    saver = tf.train.Saver()

    # Start session
    with model.session as sess:
        trainIter = 0
        valIter = 0
        bestVal = 0
        writeSummary = False

        # If we are not testing for convergence, write summaries
        if not testConvergence:
            train_writer = tf.summary.FileWriter(logs_path + '/train/'+modelName, graph=sess.graph)
            val_writer = tf.summary.FileWriter(logs_path + '/validate/'+modelName)
            writeSummary =True

        # Point to training data
        feedTrainFolder = '/home/ekremm/Documents/dataset/Video Frames/Separated by day/Every frame/Spatial/train/training/feeding'
        nonfeedTrainFolder = '/home/ekremm/Documents/dataset/Video Frames/Separated by day/Every frame/Spatial/train/training/nonfeeding'

        # Create a data handler for training data
        trainingFileHandler = File_Handler(feed_folder_path=feedTrainFolder, dataType='Spatial',
                                           nonfeed_folder_path=nonfeedTrainFolder, numTimesteps=0, datasetType=0)

        # Point to validation data
        feedValFolder = '/home/ekremm/Documents/dataset/Video Frames/Separated by day/Every frame/Spatial/train/validating2/feeding'
        nonfeedValFolder = '/home/ekremm/Documents/dataset/Video Frames/Separated by day/Every frame/Spatial/train/validating2/nonfeeding'

        # Create validation File handler
        validatingFileHandler = File_Handler(feed_folder_path=feedValFolder, dataType='Spatial',
                                             nonfeed_folder_path=nonfeedValFolder, numTimesteps=0, datasetType=0)
        # Run epochs
        for epoch_number in range(numEpochs):

            # Turn on all training variables
            tflearn.is_training(True, session=sess)


            train_startPos = 0
            trainCompImage = np.zeros(shape=(224, 224, 3))


            # Shuffle training data
            trainingFileHandler.shuffle_GRAY_data()

            # If we are testing convergence, select same batch each time
            if testConvergence:
                batch_x, batch_y, train_startPos = trainingFileHandler.get_GRAY_Batch(batch_size, train_startPos)
                #batch_x, batch_y, train_startPos = trainingFileHandler.get_GrayUnprocessed_Batch(batch_size, train_startPos)

            # Run mini-batches
            for step in range((len(trainingFileHandler.__original_feed_list)*2)/batch_size):

                # Get new batch
                if not testConvergence:
                    batch_x, batch_y, train_startPos = trainingFileHandler.get_GRAY_Batch(batch_size, train_startPos)
                    #batch_x, batch_y, train_startPos = trainingFileHandler.get_GrayUnprocessed_Batch(batch_size, train_startPos)


                # Do check to see if dataset is compromised
                if np.array_equal(trainCompImage, batch_x[0, :, :, :]):
                    print "Train images are equal!! TRAIN SET IS FUCKED!"
                trainCompImage = batch_x[0, :, :, :]


                # Do forward and backward pass with mini-batch, updating the weights
                _, loss_train, acc_train, pred_train, trainSummary = sess.run([train_op, loss, accuracy, network, merged_summary_op],
                                                                              feed_dict={x: batch_x, y_: batch_y})

                # Print results from forward pass
                print("Epoch: " + str(epoch_number) \
                      + ", Step: " + str(trainIter) \
                      + " --Loss: " + str("%.4f" % round(loss_train, 2)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_train, 2)) \
                      + "--   --" + str(step * batch_size) \
                      + "/" + str(len(trainingFileHandler.__original_feed_list) * 2))

                # Add to summary
                if writeSummary:
                    train_writer.add_summary(trainSummary, trainIter)

                trainIter += 1

                # Print comparrison between net output and true labels
                if (trainIter%100) == 0:
                    print "Accuracy: " + str(acc_train)
                    print 'Correct \t\tPred'
                    for n, g in zip(batch_y, pred_train):
                        print str(n) + '\t\t\t' + str(expit(g))


            # Prepare for validation
            val_startPos = 0
            totalVal = 0
            valCompImage = np.zeros(shape=(224, 224, 3))

            # Turn off all training variables
            tflearn.is_training(False, session=sess)

            # Run all mini-batches in validation data
            for val_step in range(len(validatingFileHandler.__original_feed_list*2)/(batch_size)):

                # Get new batch
                val_batch_x, val_batch_y, val_startPos = validatingFileHandler.get_GRAY_Batch(batch_size, val_startPos)
                #val_batch_x, val_batch_y, val_startPos = validatingFileHandler.get_GrayUnprocessed_Batch(batch_size, val_startPos)


                # Do check to see if dataset is compromised
                if np.array_equal(valCompImage, val_batch_x[0, :, :, :]):
                    print "Val images are equal!! VALIDATIONS SET IS FUCKED!"
                valCompImage = val_batch_x[0, :, :, :]

                # Perform forward pass WITHOUT the backward pass
                loss_val, acc_val, pred_val, valSummary = sess.run([loss, accuracy, network, merged_summary_op],
                                                                   feed_dict={x: val_batch_x, y_: val_batch_y})
                # Write summary
                if writeSummary:
                    val_writer.add_summary(valSummary, valIter)


                totalVal += acc_val

                # Print results from forward pass
                print("---- VALIDATION -------"
                      + " Step: " + str(valIter) \
                      + " --Loss: " + str("%.4f" % round(loss_val, 2)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_val, 2)) \
                      + "--   --" + str(val_step * batch_size) \
                      + "/" + str((len(validatingFileHandler.__original_feed_list) * 2)))

                # Print comparrison between net output and true labels
                if val_step%100 == 0:
                    print "Accuracy: " + str(acc_val)
                    print 'Correct \t\tPred'
                    for n, g in zip(val_batch_y, pred_val):
                        print str(n) + '\t\t\t' + str(expit(g))

                valIter +=1

            # Calculate total validation performance
            totalValidation = float(totalVal) / float(val_step + 1)
            VALIDATION = str("%.4f" % round(totalValidation, 4))

            print "Validation accuracy: " + VALIDATION

            # If new best validation results, save model
            if VALIDATION > bestVal:
                bestVal = VALIDATION
                # Save model
                saver.save(sess, "Models/Spatial/fineTune_"+modelName+"_step" + str(trainIter) + "_" \
                           + strftime("%Y-%m-%d_", gmtime()) + "_Val_" + str(bestVal), global_step=trainIter)


        print bestVal


#trainSpatialStream()



