import tflearn
import numpy as np
import tensorflow as tf
from tflearn.layers.core import input_data, dropout, fully_connected
from Temporal_File_Handler import File_Handler
from time import gmtime, strftime
from scipy.special import expit
from matplotlib import pyplot as plt



datasetType = 'frame'
numTimesteps = 10
tensorBoardInterval = 1


def trainMultiStream():
    """

    Trains a MultiStream 3D-CNN from scratch.


    1. Define Network Architecture
    2. Define loss and optimizer
    3. Load dataset
    4. Start training
    5. Store best performing model


    :return: Nothing
    """






    # Input placeholders
    x = tf.placeholder(shape=(None, numTimesteps, 224, 224, 3), dtype=tf.float32, name='X_placeholder')
    y_ = tf.placeholder(shape=(None, 1),  dtype=tf.float32, name='Y_placeholder')

    # Network input
    network = tflearn.input_data(placeholder=x, name='Net_in')

    # Stream1
    stream1 = tflearn.layers.conv.conv_3d(network, 23, [7, 7, 3], activation='relu', name='Conv1_1')
    stream1 = tflearn.max_pool_3d(stream1, [1, 1, 2, 2, 1], strides=[1, 2, 2])

    # Stream2
    stream2 = tflearn.layers.conv.conv_3d(network, 23, [7, 7, 5], activation='relu', name='Conv2_1')
    stream2 = tflearn.max_pool_3d(stream2, [1, 1, 2, 2, 1], strides=[1, 2, 2])

    # Stream1_x
    stream1_1 = tflearn.layers.conv.conv_3d(stream1, 13, [7, 7, 3], activation='relu', name='Conv1_2_1')
    stream1_2 = tflearn.layers.conv.conv_3d(stream1, 13, [5, 5, 3], activation='relu', name='Conv1_2_2')
    stream1_3 = tflearn.layers.conv.conv_3d(stream1, 13, [3, 3, 3], activation='relu', name='Conv1_2_3')

    # Stream2_x
    stream2_1 = tflearn.layers.conv.conv_3d(stream2, 13, [7, 7, 3], activation='relu', name='Conv2_2_1')
    stream2_2 = tflearn.layers.conv.conv_3d(stream2, 13, [5, 5, 3], activation='relu', name='Conv2_2_2')
    stream2_3 = tflearn.layers.conv.conv_3d(stream2, 13, [3, 3, 3], activation='relu', name='Conv2_2_3')

    # Pooling
    stream1_1 = tflearn.max_pool_3d(stream1_1, 2, strides=2)
    stream1_2 = tflearn.max_pool_3d(stream1_2, 3, strides=2)
    stream1_3 = tflearn.max_pool_3d(stream1_3, 3, strides=2)

    stream2_1 = tflearn.max_pool_3d(stream2_1, 2, strides=2)
    stream2_2 = tflearn.max_pool_3d(stream2_2, 3, strides=2)
    stream2_3 = tflearn.max_pool_3d(stream2_3, 3, strides=2)

    # Merge streams
    network = tflearn.merge_outputs([stream1_1,stream1_2, stream1_3, stream2_1, stream2_2, stream2_3], name='Merge')

    # Fully connected layer
    network = fully_connected(network, 256, activation='relu', name='FC1')

    # Dropout for training
    network = dropout(network, 0.5)

    # Output
    network = tflearn.fully_connected(network, 1, name='Net_out')


    # Calculate loss
    loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=y_, logits=network))

    # Set optimizer
    train_op = tf.train.AdamOptimizer(learning_rate=0.001, epsilon=0.00001).minimize(loss)

    # Calculate accuracy
    correct_prediction = tf.equal(tf.round(tf.nn.sigmoid(network)), y_)
    accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))





    # Create a summary to monitor cost tensor
    tf.summary.scalar("loss", loss)
    # Create a summary to monitor accuracy tensor
    tf.summary.scalar("accuracy", accuracy)
    merged_summary_op = tf.summary.merge_all()

    # Define number of epochs we want to run
    numEpochs = 9
    batch_size = 10

    # Initializing the variables
    init = tf.global_variables_initializer()
    logs_path = 'logs/Multi_Stream'
    trainCompImage = np.zeros(shape=(224,224,3))
    valCompImage = np.zeros(shape=(224,224,3))

    saver = tf.train.Saver()
    modelname = strftime("%Y-%m-%d-%H-%M", gmtime())+ "_KeepTimeDim(1, 0)"

    # Start session
    with tf.Session() as sess:
        sess.run(init)
        trainIter = 0
        valIter =0
        train_writer = tf.summary.FileWriter(logs_path + '/train/'+modelname, graph=sess.graph)
        val_writer = tf.summary.FileWriter(logs_path + '/validate/'+modelname, graph=sess.graph)

        bestVal = 0

        # Run epochs
        for epoch_number in range(numEpochs):
            train_startPos = 0
            # Turn on training parameters
            tflearn.is_training(True, session=sess)

            # Point to training data
            feedTrainFolder = '/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every ' + datasetType + '/Optical Flow/train/training/feeding'
            nonfeedTrainFolder = '/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every ' + datasetType + '/Optical Flow/train/training/nonfeeding'

            # Create a data handler for training data
            trainingFileHandler = File_Handler(feedTrainFolder, nonfeedTrainFolder, numTimesteps, datasetType)

            # Shuffle training data
            trainingFileHandler.shuffle_data()

            # Run mini-batches
            for step in range(len(trainingFileHandler.timeStep_nonfeed_list)*2/(batch_size)):

                # Get mini-batch
                batch_x, batch_y, train_startPos = trainingFileHandler.get_Batch(batch_size, train_startPos)

                # Check if our dataset is broken
                if np.array_equal(trainCompImage, batch_x[0, 0, :, :, :]):
                    print "Train images are equal!! TRAIN SET IS FUCKED!"


                trainCompImage = batch_x[0, 0, :, :, :]

                # Do forward and backward pass with mini-batch, updating the weights
                _, loss_train, acc_train, pred_train, trainSummary = sess.run([train_op, loss, accuracy, network, merged_summary_op],
                                                                              feed_dict={x: batch_x, y_: batch_y})

                # Add to summary
                if trainIter%tensorBoardInterval==0:
                    train_writer.add_summary(trainSummary, trainIter)

                # Print results from forward pass
                print("Epoch: " + str(epoch_number) \
                      + ", Step: " + str(trainIter) \
                      + " --Loss: " + str("%.4f" % round(loss_train, 2)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_train, 2)) \
                      + "--   --" + str(step * batch_size) \
                      + "/" + str((len(trainingFileHandler.timeStep_feed_list)+len(trainingFileHandler.timeStep_nonfeed_list))))
                trainIter += 1

                # Print comparrison between net output and true labels
                if (trainIter%100) == 0:
                    print "Accuracy: " + str(acc_train)
                    print 'Correct \t\tPred'
                    for n, g in zip(batch_y, pred_train):
                        print str(n) + '\t\t\t' + str(expit(g))


            # Prepare for validation
            val_startPos = 0
            totalVal = 0

            # Turn off training parameters
            tflearn.is_training(False, session=sess)

            # Point to validation data
            feedValFolder = '/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every ' + datasetType + '/Optical Flow/train/validating2/feeding'
            nonfeedValFolder = '/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every ' + datasetType + '/Optical Flow/train/validating2/nonfeeding'

            # Create validation File handler
            validatingFileHandler = File_Handler(feedValFolder, nonfeedValFolder, numTimesteps, datasetType)

            # Run all mini-batches in validation data
            for val_step in range(len(validatingFileHandler.timeStep_feed_list)*2/(batch_size)):

                # Get new batch
                val_batch_x, val_batch_y, val_startPos = validatingFileHandler.get_Batch(batch_size, val_startPos)

                # Do check to see if dataset is compromised
                if np.array_equal(valCompImage, val_batch_x[0, 0, :, :, :]):
                    print "Val images are equal!! VALIDATIONS SET IS FUCKED!"

                valCompImage = val_batch_x[0, 0, :, :, :]

                # Perform forward pass WITHOUT the backward pass
                loss_val, acc_val, pred_val, valSummary = sess.run([loss, accuracy, network, merged_summary_op],
                                                                feed_dict={x: val_batch_x, y_: val_batch_y})

                # Add to summary
                if valIter%tensorBoardInterval == 0:
                    val_writer.add_summary(valSummary, valIter)

                totalVal += acc_val

                # Print results from forward pass
                print("---- VALIDATION -------"
                      + " Step: " + str(valIter) \
                      + " --Loss: " + str("%.4f" % round(loss_val, 2)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_val, 2)) \
                      + "--   --" + str(val_step * batch_size) \
                      + "/" + str((len(validatingFileHandler.timeStep_feed_list) + len(validatingFileHandler.timeStep_nonfeed_list))))

                # Print comparrison between net output and true labels
                if val_step%100 == 0:
                    print "Accuracy: " + str(acc_val)
                    print 'Correct \t\tPred'
                    for n, g in zip(val_batch_y, pred_val):
                        print str(n) + '\t\t\t' + str(expit(g))

                valIter +=1

            # Calculate total validation performance
            totalValidation = float(totalVal) / float(val_step + 1)
            VALIDATION = str("%.4f" % round(totalValidation, 4))
            print "Validation accuracy: " + VALIDATION

            # If new best validation results, save model
            if bestVal < totalValidation:
                bestVal = totalValidation
                # Save model
                save_path = saver.save(sess, "Models/Multi_Stream/Multi_Stream_" + modelname + "_step_" + str(trainIter) + "_Val_"
                                       + VALIDATION + ".ckpt")

        print bestVal
#trainMultiStream()