

#Inspired by https://arxiv.org/abs/1412.0767, a twin-stream network is proposed which contains both the 3x3x3 filters
#and the 3x3x(increasing) filters which increase the temporal depth with network depth.



import tflearn
import numpy as np
import tensorflow as tf
from tflearn.layers.core import input_data, dropout, fully_connected
from Temporal_File_Handler import File_Handler
from time import gmtime, strftime
from scipy.special import expit

datasetType = 'frame'
numTimesteps = 10
tensorBoardInterval = 1








def trainTwinStream():
    """

    Trains a TwinStream 3D-CNN from scratch.


    1. Define Network Architecture
    2. Define loss and optimizer
    3. Load dataset
    4. Start training
    5. Store best performing model


    :return: Nothing
    """

    # Input placeholders
    x = tf.placeholder(shape=(None, numTimesteps, 224, 224, 3), dtype=tf.float32, name='X_placeholder')
    y_ = tf.placeholder(shape=(None, 1),  dtype=tf.float32, name='Y_placeholder')

    # Network input
    network = tflearn.input_data(placeholder=x, name='Net_in')

    # Stream 1
    stream1 = tflearn.batch_normalization(network)
    stream1 = tflearn.layers.conv.conv_3d(stream1, 32, [3, 3, 3], activation='relu', name='Stream1_Conv1')
    # Keep temporal depth
    stream1 = tflearn.max_pool_3d(stream1, [1, 1, 2, 2, 1], strides=[1, 2, 2])


    stream1 = tflearn.batch_normalization(stream1)
    stream1 = tflearn.layers.conv.conv_3d(stream1, 64, [3, 3, 3], activation='relu', name='Stream1_Conv2')
    stream1 = tflearn.max_pool_3d(stream1, 2, strides=2)

    stream1 = tflearn.batch_normalization(stream1)
    stream1 = tflearn.layers.conv.conv_3d(stream1, 128, [3, 3, 3], activation='relu', name='Stream1_Conv3')
    stream1 = tflearn.max_pool_3d(stream1, 2, strides=2)

    stream1 = tflearn.batch_normalization(stream1)
    stream1 = tflearn.layers.conv.conv_3d(stream1, 128, [3, 3, 3], activation='relu', name='Stream1_Conv4')
    stream1 = tflearn.max_pool_3d(stream1, 2, strides=2)

    stream1 = tflearn.batch_normalization(stream1)
    stream1 = tflearn.layers.conv.conv_3d(stream1, 128, [3, 3, 3], activation='relu', name='Stream1_Conv5')
    stream1 = tflearn.max_pool_3d(stream1, 2, strides=2)

    # Steam 2
    stream2 = tflearn.batch_normalization(network)
    stream2 = tflearn.layers.conv.conv_3d(stream2, 32, [3, 3, 3], activation='relu', name='Stream2_Conv1')
    # Keep temporal depth
    stream2 = tflearn.max_pool_3d(stream2, [1, 1, 2, 2, 1], strides=[1, 2, 2])

    stream2 = tflearn.batch_normalization(stream2)
    stream2 = tflearn.layers.conv.conv_3d(stream2, 64, [3, 3, 3], activation='relu', name='Stream2_Conv2')
    stream2 = tflearn.max_pool_3d(stream2, 2, strides=2)

    stream2 = tflearn.batch_normalization(stream2)
    stream2 = tflearn.layers.conv.conv_3d(stream2, 128, [3, 3, 5], activation='relu', name='Stream2_Conv3')
    stream2 = tflearn.max_pool_3d(stream2, 2, strides=2)

    stream2 = tflearn.batch_normalization(stream2)
    stream2 = tflearn.layers.conv.conv_3d(stream2, 128, [3, 3, 5], activation='relu', name='Stream2_Conv4')
    stream2 = tflearn.max_pool_3d(stream2, 2, strides=2)

    stream2 = tflearn.batch_normalization(stream2)
    stream2 = tflearn.layers.conv.conv_3d(stream2, 128, [3, 3, 7], activation='relu', name='Stream2_Conv5')
    stream2 = tflearn.max_pool_3d(stream2, 2, strides=2)

    # Merge streams
    network = tflearn.merge_outputs([stream1, stream2])

    # Fully connected
    network = tflearn.batch_normalization(network)
    network = fully_connected(network, 2048, activation='relu', name='FC1')

    network = tflearn.batch_normalization(network)
    network = fully_connected(network, 2048, activation='relu', name='FC2')

    # Output
    network = tflearn.batch_normalization(network)
    network = tflearn.fully_connected(network, 1, name='Net_out')


    # Calculate loss
    loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=y_, logits=network))

    learing_rate = 0.0001
    epsilon = 1e-8
    testConvergence = False


    # Set optimizer
    train_op = tf.train.AdamOptimizer(learning_rate=learing_rate, epsilon=epsilon).minimize(loss)

    # Calculate accuracy
    correct_prediction = tf.equal(tf.round(tf.nn.sigmoid(network)), y_)
    accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))




    # Create a summary to monitor cost tensor
    tf.summary.scalar("loss", loss)
    # Create a summary to monitor accuracy tensor
    tf.summary.scalar("accuracy", accuracy)
    merged_summary_op = tf.summary.merge_all()

    # Define number of epochs we want to run
    numEpochs = 9
    batch_size = 10

    # Initializing the variables
    init = tf.global_variables_initializer()
    logs_path = 'logs/Twin_Stream'
    trainCompImage = np.zeros(shape=(224,224,3))
    valCompImage = np.zeros(shape=(224,224,3))

    saver = tf.train.Saver()
    modelname= strftime("%Y-%m-%d-%H-%M", gmtime())+'TWIN_5conv(3FC)_KeepTimeDim(1, 0, 0, 0, 0)'

    writeSummary = False
    continueTraining = False

    # Start session
    with tf.Session() as sess:

        # If we are training from scratch
        if not continueTraining:
            print "Training model from scratch"
            sess.run(init)
            trainIter = 0
            valIter = 0
            bestVal = 0

        # If we are loading a previously trained model, prepare some parameters and load model
        else:
            print "Loading pretrained model"
            path_to_model = "/home/haakon/PycharmProjects/DNN-Masters-code/3DConv/Models/ResNet/09:54:25_25-04-2017_resNet101_type:B_KeepTime@(0, 0, 0, 0)_NumTimesteps:10_BatchSize:10_LearningRate:0.0001_Epsilon:1e-10_Val_0.7769_step-6532"
            saver.restore(sess, path_to_model)
            numEpochs = 5
            if "ckpt" in path_to_model:
                bestVal = path_to_model[-11:]
                bestVal = bestVal[:6]
                bestVal = float(bestVal)
            else:
                bestVal = path_to_model[-16:]
                bestVal = bestVal[:6]
                bestVal = float(bestVal)

            print type(bestVal)
            print bestVal
            train_op = tf.get_collection("train_op")[0]
            modelname = modelname + "_CONTINUED TRAINING"
            trainIter = 6532
            valIter = 1605


        # If we are not testing for convergence, we would like to save the statistics from the training to tensorboard
        if not testConvergence:
            train_writer = tf.summary.FileWriter(logs_path + '/train/'+modelname, graph=sess.graph)
            val_writer = tf.summary.FileWriter(logs_path + '/validate/'+modelname, graph=sess.graph)
            writeSummary =True

        # Point to training data
        feedTrainFolder = '/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every ' + \
                          datasetType + '/Optical Flow/train/training/feeding'
        nonfeedTrainFolder = '/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every ' + \
                             datasetType + '/Optical Flow/train/training/nonfeeding'

        # Create a data handler for training
        trainingFileHandler = File_Handler(feedTrainFolder, nonfeedTrainFolder, numTimesteps, datasetType)

        # Point to validation data
        feedValFolder = '/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every ' \
                        + datasetType + '/Optical Flow/train/validating/feeding'
        nonfeedValFolder = '/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every ' \
                           + datasetType + '/Optical Flow/train/validating/nonfeeding'

        # Create a data handler for validation
        validatingFileHandler = File_Handler(feedValFolder, nonfeedValFolder, numTimesteps, datasetType)

        # Run epochs
        for epoch_number in range(numEpochs):
            # Turn on training parameters
            tflearn.is_training(True, session=sess)
            train_startPos = 0

            # Shuffle training data
            trainingFileHandler.shuffle_data()


            if testConvergence:
                batch_x, batch_y, train_startPos = trainingFileHandler.get_Batch(batch_size, train_startPos)

            # Run mini-batches
            for step in range(len(trainingFileHandler.timeStep_nonfeed_list)*2/(batch_size)):
                if not testConvergence:
                    batch_x, batch_y, train_startPos = trainingFileHandler.get_Batch(batch_size, train_startPos)


                # Check if our dataset is broken
                if np.array_equal(trainCompImage, batch_x[0, 0, :, :, :]):
                    print "Train images are equal!! TRAIN SET IS FUCKED!"

                trainCompImage = batch_x[0, 0, :, :, :]

                # Do forward and backward pass with mini-batch, updating the weights
                _, loss_train, acc_train, pred_train, trainSummary = sess.run([train_op, loss, accuracy, network,
                                                                               merged_summary_op],
                                                                              feed_dict={x: batch_x, y_: batch_y})

                # Add to summary
                if trainIter%tensorBoardInterval==0 and writeSummary:
                    train_writer.add_summary(trainSummary, trainIter)

                print("Epoch: " + str(epoch_number) \
                      + ", Step: " + str(trainIter) \
                      + " --Loss: " + str("%.4f" % round(loss_train, 2)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_train, 2)) \
                      + "--   --" + str(step * batch_size) \
                      + "/" + str((len(trainingFileHandler.timeStep_feed_list)+
                                   len(trainingFileHandler.timeStep_nonfeed_list))))
                trainIter += 1

                # Print comparrison between net output and true label
                if (trainIter%10) == 0:
                    print "Accuracy: " + str(acc_train)
                    print 'Correct \t\tPred'
                    for n, g in zip(batch_y, pred_train):
                        print str(n) + '\t\t\t' + str(expit(g))

            # Prepare for validation
            val_startPos = 0
            totalVal = 0

            # Turn off training parameters
            tflearn.is_training(False, session=sess)

            # Run all mini-batches in validation data
            for val_step in range(len(validatingFileHandler.timeStep_feed_list)*2/(batch_size)):

                # Get new batch
                val_batch_x, val_batch_y, val_startPos = validatingFileHandler.get_Batch(batch_size, val_startPos)

                # Do check to see if dataset is compromised
                if np.array_equal(valCompImage, val_batch_x[0, 0, :, :, :]):
                    print "Val images are equal!! VALIDATIONS SET IS FUCKED!"

                valCompImage = val_batch_x[0, 0, :, :, :]

                # Perform forward pass WITHOUT the backward pass
                loss_val, acc_val, pred_val, valSummary = sess.run([loss, accuracy, network, merged_summary_op],
                                                                feed_dict={x: val_batch_x, y_: val_batch_y})
                # Add to summary
                if valIter%tensorBoardInterval == 0 and writeSummary:
                    val_writer.add_summary(valSummary, valIter)

                totalVal += acc_val


                # Print results from forward pass
                print("---- VALIDATION -------"
                      + " Step: " + str(valIter) \
                      + " --Loss: " + str("%.4f" % round(loss_val, 2)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_val, 2)) \
                      + "--   --" + str(val_step * batch_size) \
                      + "/" + str((len(validatingFileHandler.timeStep_feed_list) +
                                   len(validatingFileHandler.timeStep_nonfeed_list))))

                # Print comparrison between net output and true labels
                if val_step%100 == 0:
                    print "Accuracy: " + str(acc_val)
                    print 'Correct \t\tPred'
                    for n, g in zip(val_batch_y, pred_val):
                        print str(n) + '\t\t\t' + str(expit(g))

                valIter +=1

            # Calculate total validation performance
            totalValidation = float(totalVal) / float(val_step + 1)
            VALIDATION = str("%.4f" % round(totalValidation, 4))
            print "Validation accuracy: " + VALIDATION

            # If new best validation results, save model
            if bestVal < totalValidation:
                bestVal = totalValidation
                # Save model
                save_path = saver.save(sess, "Models/Twin_Stream/Twin_" + modelname + "_step_" + str(trainIter) + "_Val_"
                                       + VALIDATION)

        print bestVal
#trainTwinStream()