
#Inspired by https://arxiv.org/pdf/1512.03385.pdf


import tflearn
from ResnetLayers import residual_block3D, residual_bottleneck3D, global_avg_pool3D
from Temporal_File_Handler import File_Handler
import ResNetModels3D
import numpy as np
import tensorflow as tf
from tflearn.layers.core import input_data, dropout, fully_connected
from Temporal_File_Handler import File_Handler
from time import gmtime, strftime
from scipy.special import expit
import sys

# Change path based on computer used
if sys.path[0].__contains__('haakon'):
    startPath = "/home/haakon/Documents/Datasets/Large Dataset/"
else:
    startPath = "/home/ekremm/Documents/dataset/"

datasetType = 'frame'
numTimesteps = 20
tensorBoardInterval = 1


def train3DResNet():
    """

       Trains a 3D-Residual Neural Network from scratch.


       1. Define Network Architecture
       2. Define loss and optimizer
       3. Load dataset
       4. Start training
       5. Store best performing model


       :return: Nothing
       """

    # Input placeholders
    x = tf.placeholder(shape=(None, numTimesteps, 224, 224, 3), dtype=tf.float32, name='X_placeholder')
    y_ = tf.placeholder(shape=(None, 1),  dtype=tf.float32, name='Y_placeholder')

    # Choose which 3D-ResNet to train(choices available at ResNetModels3D.py)
    network, name, numResBlock, k = ResNetModels3D.resNet34(x, keep_time_dim=[1, 0, 0, 0], printLayer=[0, 0, 0], type='A')

    # Output
    network = tflearn.fully_connected(network, 1, name='Net_out')


    # Calculate loss
    loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=y_, logits=network))

    learing_rate = 0.0001
    epsilon = 1e-8
    testConvergence = True

    # Set optimizer
    train_op = tf.train.AdamOptimizer(learning_rate=learing_rate, epsilon=epsilon).minimize(loss)

    # Calculate accuracy
    correct_prediction = tf.equal(tf.round(tf.nn.sigmoid(network)), y_)
    accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))

    # Create a summary to monitor cost tensor
    tf.summary.scalar("loss", loss)
    # Create a summary to monitor accuracy tensor
    tf.summary.scalar("accuracy", accuracy)
    merged_summary_op = tf.summary.merge_all()

    # Define number of epochs we want to run
    numEpochs = statsFile.numEpochs
    batch_size = statsFile.batchSize

    # Initializing the variables
    init = tf.global_variables_initializer()
    logs_path = 'logs/ResNet18'
    trainCompImage = np.zeros(shape=(224,224,3))
    valCompImage = np.zeros(shape=(224,224,3))

    # Naming of network
    if numResBlock == 0:
        modelname = strftime("%H:%M:%S_%d-%m-%Y", gmtime())+"_"+name+"_NumTimesteps:"+str(numTimesteps)+\
                    "_BatchSize:"+str(batch_size)+ "_LearningRate:"+str(learing_rate)+"_Epsilon:"+str(epsilon)
    else:
        modelname = strftime("%H:%M:%S_%d-%m-%Y", gmtime())+"_"+name + "_NumTimesteps:" + str(numTimesteps) +\
                    "_BatchSize:" + str(batch_size) + "_LearningRate:"+str(learing_rate)+"_Epsilon:"+str(epsilon)\
                    +"_NumResblocks:"+str(numResBlock)+ "_k:"+str(k)

    saver = tf.train.Saver()
    writeSummary = False
    continueTraining = False

    # Start session
    with tf.Session() as sess:

        # If we are training from scratch
        if not continueTraining:
            print "Training model from scratch"
            sess.run(init)
            trainIter = 0
            valIter = 0
            bestVal = 0

        # If we are loading a previously trained model, prepare some parameters and load model
        else:
            print "Loading pretrained model"
            path_to_model = "/home/haakon/PycharmProjects/DNN-Masters-code/Three_DConv/Models/ResNet18/22:13:52_04-06-2017_resNet32_type:A_KeepTime@(1, 0, 0, 0)_NumTimesteps:20_BatchSize:10_LearningRate:0.0001_Epsilon:1e-08_CONTINUED-TRAINING_Every-Frame_NOAugmented_Val_0.8371_step-1633"
            saver.restore(sess, path_to_model)
            numEpochs = 5
            if "ckpt" in path_to_model:
                bestVal = path_to_model[-11:]
                bestVal = bestVal[:6]
                bestVal = float(bestVal)
            else:
                bestVal = path_to_model[-16:]
                bestVal = bestVal[:6]
                bestVal = float(bestVal)

            print type(bestVal)
            print bestVal
            #train_op = tf.get_collection("train_op")[0]
            modelname = modelname + "_CONTINUED-TRAINING"
            trainIter = 0
            valIter = 0




        # Training using every second frame
        if datasetType == "2nd":
            print "Every 2nd frame"
            # Pont to training data
            feedTrainFolder = startPath+'Video Frames/Separated by day/Every ' + \
                              datasetType + '/Optical Flow/Original/train/training/feeding'
            nonfeedTrainFolder = startPath+'Video Frames/Separated by day/Every ' + \
                                 datasetType + '/Optical Flow/Original/train/training/nonfeeding'


            # Create a data handler for training
            trainingFileHandler = File_Handler(feedTrainFolder, nonfeedTrainFolder, numTimesteps, datasetType)

            # Pint to validation data
            feedValFolder = startPath+'Video Frames/Separated by day/Every ' \
                            + datasetType + '/Optical Flow/Original/train/validating2/feeding'
            nonfeedValFolder = startPath+'Video Frames/Separated by day/Every ' \
                               + datasetType + '/Optical Flow/Original/train/validating2/nonfeeding'

            # Create data handler for validation
            validatingFileHandler = File_Handler(feedValFolder, nonfeedValFolder, numTimesteps, datasetType)

            # Add sampling rate to model name
            modelname = modelname + "_Every-2nd"

        # Training using every frame
        elif datasetType == "frame":
            print "Every frame"
            # Pont to training data
            feedTrainFolder = "/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every frame/Optical Flow/Original/train/training/feeding"
            nonfeedTrainFolder = "/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every frame/Optical Flow/Original/train/training/nonfeeding"

            # Create a data handler for training
            trainingFileHandler = File_Handler(feedTrainFolder, nonfeedTrainFolder, numTimesteps, datasetType)

            # Pint to validation data
            feedValFolder = "/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every frame/Optical Flow/Augmented/train/validating2/feeding"
            nonfeedValFolder = "/home/haakon/Documents/Datasets/Large Dataset/Video Frames/Separated by day/Every frame/Optical Flow/Augmented/train/validating2/nonfeeding"

            # Create a data handler for validation
            validatingFileHandler = File_Handler(feedValFolder, nonfeedValFolder, numTimesteps, datasetType)

            # Add sampling rate to model name
            modelname = modelname + "_Every-Frame"

        # If we are not testing for convergence, we would like to save the statistics from the training to tensorboard
        if not testConvergence:
            modelname = modelname
            train_writer = tf.summary.FileWriter(logs_path + '/train/' + modelname, graph=sess.graph)
            val_writer = tf.summary.FileWriter(logs_path + '/validate/' + modelname, graph=sess.graph)
            writeSummary = True

        # Run epochs
        for epoch_number in range(numEpochs):
            # Turn on training parameters
            tflearn.is_training(True, session=sess)
            train_startPos = 0

            # Shuffle training data
            trainingFileHandler.shuffle_data()

            if testConvergence:
                batch_x, batch_y, train_startPos = trainingFileHandler.get_Batch(batch_size, train_startPos)

            # Run mini-batches
            for step in range(len(trainingFileHandler.timeStep_nonfeed_list)*2/(batch_size)):


                if not testConvergence:
                    # Get mini-batch
                    batch_x, batch_y, train_startPos = trainingFileHandler.get_Batch(batch_size, train_startPos)


                # Check if our dataset is broken
                if np.array_equal(trainCompImage, batch_x[0, 0, :, :, :]):
                    print "Train images are equal!! TRAIN SET IS FUCKED!"

                trainCompImage = batch_x[0, 0, :, :, :]

                ## Do forward and backward pass with mini-batch, updating the weights
                _, loss_train, acc_train, pred_train, trainSummary = sess.run([train_op, loss, accuracy, network,
                                                                               merged_summary_op],
                                                                              feed_dict={x: batch_x, y_: batch_y})

                # Add to summary
                if trainIter%tensorBoardInterval==0 and writeSummary:
                    train_writer.add_summary(trainSummary, trainIter)

                # Print results from forward pass
                print("Epoch: " + str(epoch_number) \
                      + ", Step: " + str(trainIter) \
                      + " --Loss: " + str("%.4f" % round(loss_train, 2)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_train, 2)) \
                      + "--   --" + str(step * batch_size) \
                      + "/" + str((len(trainingFileHandler.timeStep_feed_list)+
                                   len(trainingFileHandler.timeStep_nonfeed_list))))
                trainIter += 1

                # Print comparrison between net output and true labels
                if (trainIter%10) == 0:
                    print "Accuracy: " + str(acc_train)
                    print 'Correct \t\tPred'
                    for n, g in zip(batch_y, pred_train):
                        print str(n) + '\t\t\t' + str(expit(g))

            # Prepare for validation
            val_startPos = 0
            totalVal = 0

            # Turn off training parameters
            tflearn.is_training(False, session=sess)

            # Run all mini-batches in validation data
            for val_step in range(len(validatingFileHandler.timeStep_feed_list)*2/(batch_size)):
                # Get new batch
                val_batch_x, val_batch_y, val_startPos = validatingFileHandler.get_Batch(batch_size, val_startPos)

                # Do check to see if dataset is compromised
                if np.array_equal(valCompImage, val_batch_x[0, 0, :, :, :]):
                    print "Val images are equal!! VALIDATIONS SET IS FUCKED!"

                valCompImage = val_batch_x[0, 0, :, :, :]

                # Perform forward pass WITHOUT the backward pass
                loss_val, acc_val, pred_val, valSummary = sess.run([loss, accuracy, network, merged_summary_op],
                                                                feed_dict={x: val_batch_x, y_: val_batch_y})

                # Add to summary
                if valIter%tensorBoardInterval == 0 and writeSummary:
                    val_writer.add_summary(valSummary, valIter)

                totalVal += acc_val

                # Print results from forward pass
                print("---- VALIDATION -------"
                      + " Step: " + str(valIter) \
                      + " --Loss: " + str("%.4f" % round(loss_val, 2)) \
                      + "--  --Accuracy: " + str("%.4f" % round(acc_val, 2)) \
                      + "--   --" + str(val_step * batch_size) \
                      + "/" + str((len(validatingFileHandler.timeStep_feed_list) +
                                   len(validatingFileHandler.timeStep_nonfeed_list))))

                # Print comparrison between net output and true labels
                if val_step%100 == 0:
                    print "Accuracy: " + str(acc_val)
                    print 'Correct \t\tPred'
                    for n, g in zip(val_batch_y, pred_val):
                        print str(n) + '\t\t\t' + str(expit(g))

                valIter +=1

            # Calculate total validation performance
            totalValidation = float(totalVal) / float(val_step + 1)
            VALIDATION = str("%.4f" % round(totalValidation, 4))
            print "Validation accuracy: " + VALIDATION

            # If new best validation results, save model
            if bestVal < totalValidation:
                bestVal = totalValidation
                # Save model
                save_path = saver.save(sess, "Models/ResNet18/" +modelname + "_NOAugmented_Val_"
                                       + VALIDATION+"_step", global_step=trainIter)

        print bestVal
