import tflearn
from ResnetLayers import residual_block3D, residual_bottleneck3D, global_avg_pool3D, LongResidual_block3D, long_residual_bottleneck3D



def resNet34(x, keep_time_dim=[False, False, False, False], printLayer=[False, False, False], type='B'):
    """
    Defines a 34-layer 3D-Residual Neural Network from scratch, using 3D-Residual Blocks

    :param x: network input
    :param keep_time_dim: array of length 4. Used to select for how long we will keep the temporal dimension depth through the network. Each index in the array corresponds to a downsampling layer
    :param printLayer: array of length 3. Used to select which layers to print. Each index corresponds to a downsampling block
    :param type: Selects the downsampling type we want to use('A', 'B' or 'C')
    :return: the network globalAveragePool layer, the name of the network, number of residualBlocks, k is not used anymore
    """
    timeDimString = str(keep_time_dim)
    timeDimString = timeDimString.replace("[", "(")
    timeDimString = timeDimString.replace("]", ")")
    name = "resNet34_type:" + type + "_KeepTime@" + timeDimString
    network = tflearn.input_data(placeholder=x, name='Net_in')
    print "input" + str(network.shape)
    # Conv1
    if keep_time_dim[0] != 0:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=[7, 7, 7], strides=[1, 2, 2])
        print "conv1 " +str(network.shape)
        network = tflearn.max_pool_3d(network, kernel_size=[1, 1, 3, 3, 1], strides=[1, 2, 2])
    else:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=7, strides=2)
        print "conv1 " + str(network.shape)
        network = tflearn.max_pool_3d(network, kernel_size=3, strides=2)
    print "pool" + str(network.shape)
    # Conv2
    network = residual_block3D(network, nb_blocks=3, out_channels=64, type=type)
    print "conv2_x " + str(network.shape)
    # Conv3
    network = residual_block3D(network, nb_blocks=1, out_channels=128, downsample=True,
                               keep_time_dim=keep_time_dim[1], print_Layer=printLayer[0], type=type)
    print "Downsample" + str(network.shape)
    network = residual_block3D(network, nb_blocks=3, out_channels=128, type=type)
    print "conv3_x " + str(network.shape)
    # Conv4
    network = residual_block3D(network, nb_blocks=1, out_channels=256, downsample=True,
                               keep_time_dim=keep_time_dim[2], print_Layer=printLayer[1], type=type)
    print "Downsample" + str(network.shape)
    network = residual_block3D(network, nb_blocks=5, out_channels=256, type=type)
    print "conv4_x " + str(network.shape)
    # Conv5
    network = residual_block3D(network, nb_blocks=1, out_channels=512, downsample=True,
                               keep_time_dim=keep_time_dim[3], print_Layer=printLayer[2], type=type)
    print "Downsample" + str(network.shape)
    network = residual_block3D(network, nb_blocks=2, out_channels=512, type=type)
    print "conv5_x " + str(network.shape)

    network = tflearn.batch_normalization(network)
    network = tflearn.activation(network, 'relu')
    network = global_avg_pool3D(network)

    # Used for name conventions, have no function here
    num_resBlocks = 0
    k = 0

    return network, name, num_resBlocks, k


def resNet18(x, keep_time_dim=[False, False, False, False], printLayer=[False, False, False], type='B'):
    """
    Defines a 18-layer 3D-Residual Neural Network from scratch, using 3D-Residual Blocks

    :param x: network input
    :param keep_time_dim: array of length 4. Used to select for how long we will keep the temporal dimension depth through the network. Each index in the array corresponds to a downsampling layer
    :param printLayer: array of length 3. Used to select which layers to print. Each index corresponds to a downsampling block
    :param type: Selects the downsampling type we want to use('A', 'B' or 'C')
    :return: the network globalAveragePool layer, the name of the network, number of residualBlocks, k is not used anymore
    """

    timeDimString = str(keep_time_dim)
    timeDimString = timeDimString.replace("[", "(")
    timeDimString = timeDimString.replace("]", ")")
    name = "resNet18_type:"+type+"_KeepTime@"+timeDimString

    network = tflearn.input_data(placeholder=x, name='Net_in')

    if keep_time_dim[0] != 0:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=[7, 7, 7], strides=[1, 2, 2])
        network = tflearn.max_pool_3d(network, kernel_size=[1, 1, 2, 2, 1], strides=[1, 2, 2])
    else:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=7, strides=2)
        network = tflearn.max_pool_3d(network, kernel_size=3, strides=2)

    # Conv2
    network = residual_block3D(network, nb_blocks=2, out_channels=64, type=type)

    # Conv3
    network = residual_block3D(network, nb_blocks=1, out_channels=128, downsample=True,
                               keep_time_dim=keep_time_dim[1],  print_Layer=printLayer[0], type=type)
    network = residual_block3D(network, nb_blocks=1, out_channels=128, type=type)

    # Conv4
    network = residual_block3D(network, nb_blocks=1, out_channels=256, downsample=True,
                               keep_time_dim=keep_time_dim[2],  print_Layer=printLayer[1], type=type)
    network = residual_block3D(network, nb_blocks=1, out_channels=256, type=type)

    # Conv5
    network = residual_block3D(network, nb_blocks=1, out_channels=512, downsample=True,
                               keep_time_dim=keep_time_dim[3],  print_Layer=printLayer[2], type=type)
    network = residual_block3D(network, nb_blocks=1, out_channels=512, type=type)

    network = tflearn.batch_normalization(network)
    network = tflearn.activation(network, 'relu')
    network = global_avg_pool3D(network)

    # Used for name conventions, have no function here
    num_resBlocks = 0
    k = 0

    return network, name, num_resBlocks, k




def resNet152(x, keep_time_dim=[False, False, False, False], printLayer=[False, False, False], type='B'):
    """
    Defines a 152-layer 3D-Residual Neural Network from scratch, using 3D-Residual Bottlenecks

    :param x: network input
    :param keep_time_dim: array of length 4. Used to select for how long we will keep the temporal dimension depth through the network. Each index in the array corresponds to a downsampling layer
    :param printLayer: array of length 3. Used to select which layers to print. Each index corresponds to a downsampling block
    :param type: Selects the downsampling type we want to use('A', 'B' or 'C')
    :return: the network globalAveragePool layer, the name of the network, number of residualBlocks, k is not used anymore
    """

    timeDimString = str(keep_time_dim)
    timeDimString = timeDimString.replace("[", "(")
    timeDimString = timeDimString.replace("]", ")")
    name = "resNet152_type:"+type+"_KeepTime@"+timeDimString
    network = tflearn.input_data(placeholder=x, name='Net_in')

    if keep_time_dim[0] != 0:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=[7, 7, 7], strides=[1, 2, 2])
        network = tflearn.max_pool_3d(network, kernel_size=[1, 1, 2, 2, 1], strides=[1, 2, 2])
    else:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=7, strides=2)
        network = tflearn.max_pool_3d(network, kernel_size=3, strides=2)

    # Conv2
    network = residual_bottleneck3D(network, nb_blocks=3, bottleneck_size=64, out_channels=256, type=type)

    # Conv3
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=128, out_channels=512, downsample=True,
                                    keep_time_dim=keep_time_dim[1], print_Layer=printLayer[0], type=type)
    network = residual_bottleneck3D(network, nb_blocks=7, bottleneck_size=128, out_channels=512, type=type)

    # Conv4
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=256, out_channels=1024, downsample=True,
                                    keep_time_dim=keep_time_dim[2], print_Layer=printLayer[1], type=type)
    network = residual_bottleneck3D(network, nb_blocks=35, bottleneck_size=256, out_channels=1024, type=type)

    # Conv5
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=512, out_channels=2048, downsample=True,
                                    keep_time_dim=keep_time_dim[3], print_Layer=printLayer[2], type=type)
    network = residual_bottleneck3D(network, nb_blocks=2, bottleneck_size=512, out_channels=2048, type=type)

    network = tflearn.batch_normalization(network)
    network = tflearn.activation(network, 'relu')
    network = global_avg_pool3D(network)

    # Used for name conventions, have no function here
    num_resBlocks = 0
    k = 0

    return network, name, num_resBlocks, k


def resNet101(x, keep_time_dim=[False, False, False, False], printLayer=[False, False, False], type='B'):

    """
    Defines a 101-layer 3D-Residual Neural Network from scratch, using 3D-Residual Bottlenecks

    :param x: network input
    :param keep_time_dim: array of length 4. Used to select for how long we will keep the temporal dimension depth through the network. Each index in the array corresponds to a downsampling layer
    :param printLayer: array of length 3. Used to select which layers to print. Each index corresponds to a downsampling block
    :param type: Selects the downsampling type we want to use('A', 'B' or 'C')
    :return: the network globalAveragePool layer, the name of the network, number of residualBlocks, k is not used anymore
    """

    timeDimString = str(keep_time_dim)
    timeDimString = timeDimString.replace("[", "(")
    timeDimString = timeDimString.replace("]", ")")
    name = "resNet101_type:" + type + "_KeepTime@" + timeDimString
    network = tflearn.input_data(placeholder=x, name='Net_in')

    if keep_time_dim[0] != 0:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=[7, 7, 7], strides=[1, 2, 2])
        network = tflearn.max_pool_3d(network, kernel_size=[1, 1, 2, 2, 1], strides=[1, 2, 2])
    else:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=7, strides=2)
        network = tflearn.max_pool_3d(network, kernel_size=3, strides=2)

    # Conv2
    network = residual_bottleneck3D(network, nb_blocks=3, bottleneck_size=64, out_channels=256, type=type)

    # Conv3
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=128, out_channels=512, downsample=True,
                                    keep_time_dim=keep_time_dim[1], print_Layer=printLayer[0], type=type)
    network = residual_bottleneck3D(network, nb_blocks=3, bottleneck_size=128, out_channels=512, type=type)

    # Conv4
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=256, out_channels=1024, downsample=True,
                                    keep_time_dim=keep_time_dim[2], print_Layer=printLayer[1], type=type)
    network = residual_bottleneck3D(network, nb_blocks=22, bottleneck_size=256, out_channels=1024, type=type)

    # Conv5
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=512, out_channels=2048, downsample=True,
                                    keep_time_dim=keep_time_dim[3], print_Layer=printLayer[2], type=type)
    network = residual_bottleneck3D(network, nb_blocks=2, bottleneck_size=512, out_channels=2048, type=type)

    network = tflearn.batch_normalization(network)
    network = tflearn.activation(network, 'relu')
    network = global_avg_pool3D(network)

    # Used for name conventions, have no function here
    num_resBlocks = 0
    k = 0

    return network, name, num_resBlocks, k

def resNet50(x, keep_time_dim=[False, False, False, False], printLayer=[False, False, False], type='B'):
    """
    Defines a 50-layer 3D-Residual Neural Network from scratch, using 3D-Residual Bottlenecks

    :param x: network input
    :param keep_time_dim: array of length 4. Used to select for how long we will keep the temporal dimension depth through the network. Each index in the array corresponds to a downsampling layer
    :param printLayer: array of length 3. Used to select which layers to print. Each index corresponds to a downsampling block
    :param type: Selects the downsampling type we want to use('A', 'B' or 'C')
    :return: the network globalAveragePool layer, the name of the network, number of residualBlocks, k is not used anymore
    """

    timeDimString = str(keep_time_dim)
    timeDimString = timeDimString.replace("[", "(")
    timeDimString = timeDimString.replace("]", ")")
    name = "resNet50_type:" + type + "_KeepTime@" + timeDimString
    network = tflearn.input_data(placeholder=x, name='Net_in')

    if keep_time_dim[0] != 0:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=[7, 7, 7], strides=[1, 2, 2])
        network = tflearn.max_pool_3d(network, kernel_size=[1, 1, 2, 2, 1], strides=[1, 2, 2])
    else:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=7, strides=2)
        network = tflearn.max_pool_3d(network, kernel_size=3, strides=2)

    # Conv2
    network = residual_bottleneck3D(network, nb_blocks=3, bottleneck_size=64, out_channels=256, type=type)

    # Conv3
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=128, out_channels=512, downsample=True,
                                    keep_time_dim=keep_time_dim[1], print_Layer=printLayer[0], type=type)
    network = residual_bottleneck3D(network, nb_blocks=3, bottleneck_size=128, out_channels=512, type=type)

    # Conv4
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=256, out_channels=1024, downsample=True,
                                    keep_time_dim=keep_time_dim[2], print_Layer=printLayer[1], type=type)
    network = residual_bottleneck3D(network, nb_blocks=5, bottleneck_size=256, out_channels=1024, type=type)

    # Conv5
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=512, out_channels=2048, downsample=True,
                                    keep_time_dim=keep_time_dim[3], print_Layer=printLayer[2], type=type)
    network = residual_bottleneck3D(network, nb_blocks=2, bottleneck_size=512, out_channels=2048, type=type)

    network = tflearn.batch_normalization(network)
    network = tflearn.activation(network, 'relu')
    network = global_avg_pool3D(network)

    # Used for name conventions, have no function here
    num_resBlocks = 0
    k = 0

    return network, name, num_resBlocks, k


def longResNet101(x, keep_time_dim=[False, False, False, False], printLayer=[False, False, False], type='B'):
    """
    Defines a 101-layer long 3D-Residual Neural Network from scratch, using 3D-Residual Bottlenecks

    :param x: network input
    :param keep_time_dim: array of length 4. Used to select for how long we will keep the temporal dimension depth through the network. Each index in the array corresponds to a downsampling layer
    :param printLayer: array of length 3. Used to select which layers to print. Each index corresponds to a downsampling block
    :param type: Selects the downsampling type we want to use('A', 'B' or 'C')
    :return: the network globalAveragePool layer, the name of the network, number of residualBlocks, k is not used anymore
    """

    timeDimString = str(keep_time_dim)
    timeDimString = timeDimString.replace("[", "(")
    timeDimString = timeDimString.replace("]", ")")
    name = "longResNet101_type:" + type + "_KeepTime@" + timeDimString
    network = tflearn.input_data(placeholder=x, name='Net_in')

    if keep_time_dim[0] != 0:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=[7, 7, 7], strides=[1, 2, 2])
        network = tflearn.max_pool_3d(network, kernel_size=[1, 1, 2, 2, 1], strides=[1, 2, 2])
    else:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=7, strides=2)
        network = tflearn.max_pool_3d(network, kernel_size=3, strides=2)

    # Conv2
    network = long_residual_bottleneck3D(network, nb_blocks=3, bottleneck_size=64, out_channels=256, type=type)

    # Conv3
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=128, out_channels=512, downsample=True,
                                    keep_time_dim=keep_time_dim[1], print_Layer=printLayer[0], type=type)
    network = long_residual_bottleneck3D(network, nb_blocks=3, bottleneck_size=128, out_channels=512, type=type)

    # Conv4
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=256, out_channels=1024, downsample=True,
                                    keep_time_dim=keep_time_dim[2], print_Layer=printLayer[1], type=type)
    network = long_residual_bottleneck3D(network, nb_blocks=22, bottleneck_size=256, out_channels=1024, type=type)

    # Conv5
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=512, out_channels=2048, downsample=True,
                                    keep_time_dim=keep_time_dim[3], print_Layer=printLayer[2], type=type)
    network = long_residual_bottleneck3D(network, nb_blocks=2, bottleneck_size=512, out_channels=2048, type=type)

    network = tflearn.batch_normalization(network)
    network = tflearn.activation(network, 'relu')
    network = global_avg_pool3D(network)

    # Used for name conventions, have no function here
    num_resBlocks = 0
    k = 0

    return network, name, num_resBlocks, k

def longResNet50Bottle(x, keep_time_dim=[False, False, False, False], printLayer=[False, False, False], type='B'):
    """
    Defines a 50-layer long 3D-Residual Neural Network from scratch, using 3D-Residual Bottlenecks

    :param x: network input
    :param keep_time_dim: array of length 4. Used to select for how long we will keep the temporal dimension depth through the network. Each index in the array corresponds to a downsampling layer
    :param printLayer: array of length 3. Used to select which layers to print. Each index corresponds to a downsampling block
    :param type: Selects the downsampling type we want to use('A', 'B' or 'C')
    :return: the network globalAveragePool layer, the name of the network, number of residualBlocks, k is not used anymore
    """

    timeDimString = str(keep_time_dim)
    timeDimString = timeDimString.replace("[", "(")
    timeDimString = timeDimString.replace("]", ")")
    name = "longResNet50_type:" + type + "_KeepTime@" + timeDimString
    network = tflearn.input_data(placeholder=x, name='Net_in')

    if keep_time_dim[0] != 0:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=[7, 7, 7], strides=[1, 2, 2])
        network = tflearn.max_pool_3d(network, kernel_size=[1, 1, 2, 2, 1], strides=[1, 2, 2])
    else:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=7, strides=2)
        network = tflearn.max_pool_3d(network, kernel_size=3, strides=2)

    # Conv2
    network = long_residual_bottleneck3D(network, nb_blocks=3, bottleneck_size=64, out_channels=256, type=type)

    # Conv3
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=128, out_channels=512, downsample=True,
                                    keep_time_dim=keep_time_dim[1], print_Layer=printLayer[0], type=type)
    network = long_residual_bottleneck3D(network, nb_blocks=3, bottleneck_size=128, out_channels=512, type=type)

    # Conv4
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=256, out_channels=1024, downsample=True,
                                    keep_time_dim=keep_time_dim[2], print_Layer=printLayer[1], type=type)
    network = long_residual_bottleneck3D(network, nb_blocks=5, bottleneck_size=256, out_channels=1024, type=type)

    # Conv5
    network = residual_bottleneck3D(network, nb_blocks=1, bottleneck_size=512, out_channels=2048, downsample=True,
                                    keep_time_dim=keep_time_dim[3], print_Layer=printLayer[2], type=type)
    network = long_residual_bottleneck3D(network, nb_blocks=2, bottleneck_size=512, out_channels=2048, type=type)

    network = tflearn.batch_normalization(network)
    network = tflearn.activation(network, 'relu')
    network = global_avg_pool3D(network)

    # Used for name conventions, have no function here
    num_resBlocks = 0
    k = 0

    return network, name, num_resBlocks, k


def longResNet18(x, keep_time_dim=[False, False, False, False], printLayer=[False, False, False], type='B'):
    """
    Defines a 18-layer long 3D-Residual Neural Network from scratch, using 3D-Residual Blocks

    :param x: network input
    :param keep_time_dim: array of length 4. Used to select for how long we will keep the temporal dimension depth through the network. Each index in the array corresponds to a downsampling layer
    :param printLayer: array of length 3. Used to select which layers to print. Each index corresponds to a downsampling block
    :param type: Selects the downsampling type we want to use('A', 'B' or 'C')
    :return: the network globalAveragePool layer, the name of the network, number of residualBlocks, k is not used anymore
    """
    timeDimString = str(keep_time_dim)
    timeDimString = timeDimString.replace("[", "(")
    timeDimString = timeDimString.replace("]", ")")
    name = "longResNet18_type:"+type+"_KeepTime@"+timeDimString

    network = tflearn.input_data(placeholder=x, name='Net_in')

    if keep_time_dim[0] != 0:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=[7, 7, 7], strides=[1, 2, 2])
        network = tflearn.max_pool_3d(network, kernel_size=[1, 1, 2, 2, 1], strides=[1, 2, 2])
    else:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=7, strides=2)
        network = tflearn.max_pool_3d(network, kernel_size=3, strides=2)

    # Conv2
    network = LongResidual_block3D(network, nb_blocks=2, out_channels=64, type=type)

    # Conv3
    network = residual_block3D(network, nb_blocks=1, out_channels=128, downsample=True,
                               keep_time_dim=keep_time_dim[1],  print_Layer=printLayer[0], type=type)
    network = residual_block3D(network, nb_blocks=1, out_channels=128, type=type)

    # Conv4
    network = residual_block3D(network, nb_blocks=1, out_channels=256, downsample=True,
                               keep_time_dim=keep_time_dim[2],  print_Layer=printLayer[1], type=type)
    network = residual_block3D(network, nb_blocks=1, out_channels=256, type=type)

    # Conv5
    network = residual_block3D(network, nb_blocks=1, out_channels=512, downsample=True,
                               keep_time_dim=keep_time_dim[3],  print_Layer=printLayer[2], type=type)
    network = residual_block3D(network, nb_blocks=1, out_channels=512, type=type)

    network = tflearn.batch_normalization(network)
    network = tflearn.activation(network, 'relu')
    network = global_avg_pool3D(network)

    # Used for name conventions, have no function here
    num_resBlocks = 0
    k = 0

    return network, name, num_resBlocks, k


def longResNet34(x, keep_time_dim=[False, False, False, False], printLayer=[False, False, False], type='B'):
    """
    Defines a 34-layer long 3D-Residual Neural Network from scratch, using 3D-Residual Blocks

    :param x: network input
    :param keep_time_dim: array of length 4. Used to select for how long we will keep the temporal dimension depth through the network. Each index in the array corresponds to a downsampling layer
    :param printLayer: array of length 3. Used to select which layers to print. Each index corresponds to a downsampling block
    :param type: Selects the downsampling type we want to use('A', 'B' or 'C')
    :return: the network globalAveragePool layer, the name of the network, number of residualBlocks, k is not used anymore
    """
    timeDimString = str(keep_time_dim)
    timeDimString = timeDimString.replace("[", "(")
    timeDimString = timeDimString.replace("]", ")")
    name = "longResNet34_type:" + type + "_KeepTime@" + timeDimString
    network = tflearn.input_data(placeholder=x, name='Net_in')
    # Conv1
    if keep_time_dim[0] != 0:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=[7, 7, 7], strides=[1, 2, 2])
        network = tflearn.max_pool_3d(network, kernel_size=[1, 1, 2, 2, 1], strides=[1, 2, 2])
    else:
        network = tflearn.conv_3d(network, nb_filter=64, filter_size=7, strides=2)
        network = tflearn.max_pool_3d(network, kernel_size=3, strides=2)

    # Conv2
    network = LongResidual_block3D(network, nb_blocks=3, out_channels=64, type=type)

    # Conv3
    network = residual_block3D(network, nb_blocks=1, out_channels=128, downsample=True,
                               keep_time_dim=keep_time_dim[1], print_Layer=printLayer[0], type=type)
    network = LongResidual_block3D(network, nb_blocks=3, out_channels=128, type=type)

    # Conv4
    network = residual_block3D(network, nb_blocks=1, out_channels=256, downsample=True,
                               keep_time_dim=keep_time_dim[2], print_Layer=printLayer[1], type=type)
    network = LongResidual_block3D(network, nb_blocks=5, out_channels=256, type=type)

    # Conv5
    network = residual_block3D(network, nb_blocks=1, out_channels=512, downsample=True,
                               keep_time_dim=keep_time_dim[3], print_Layer=printLayer[2], type=type)
    network = LongResidual_block3D(network, nb_blocks=2, out_channels=512, type=type)

    network = tflearn.batch_normalization(network)
    network = tflearn.activation(network, 'relu')
    network = global_avg_pool3D(network)

    # Used for name conventions, have no function here
    num_resBlocks = 0
    k = 0

    return network, name, num_resBlocks, k
