import numpy as np
from matplotlib import pyplot as plt
from PIL import ImageEnhance
from skimage.exposure import equalize_hist

def zeroCenterGrayData(image):

    """
    Preprocesses image, using our Specular Removal Preprocessing strategy.




    :param image: Matrix containing an image
        Shape = [224, 224, 3]

    :return: Preprocessed image
    """

    #Images mean = 0.503778116665
    #Images std = 0.287660510217


    # Desaturate image to create grayscale
    image = ImageEnhance.Color(image)
    image = image.enhance(0.0)

    # Turn into np array for math
    image = np.asarray(image, dtype=np.float32)


    # Equalize the histogram of the image
    image[:, :, 0] = equalize_hist(image[:, :, 0])
    image[:, :, 1] = equalize_hist(image[:, :, 1])
    image[:, :, 2] = equalize_hist(image[:, :, 2])


    # Zero center image
    image -= 0.50377811666665

    # Normalize image
    image /= 0.287660510217


    # Zero out any extreme pixels, such as specular reflection from fish skin and very dark areas
    image[image < -1.6] = 0
    image[image > 1.6] = 0


    return image



def zeroCenterOFData(image, datasetType):
    """
     Normalizes and zero centers Optical Flow data

    :param image: pixel matrix (224 x 224 x 3)
    :param datasetType: the sampling rate of our data
    :return: zero-centered and normalized image, based on sampling rate used
    """


    if datasetType == 'frame':
        # Normalize image
        image /= 56.586015345
        # Zero center image
        image -= 0.605827630151

    elif(datasetType == '2nd'):
        # Normalize image
        image /= 56.5766961032
        # Zero center image
        image -= 0.605778895649

    elif(datasetType == '5th'):
        # Normalize image
        image /= 46.8160071216
        # Zero center image
        image -= 0.529848780878




    return image

